from threading import Thread
import inspect
import ctypes


# 尝试关闭线程
def stop_thread(thread):
    '''
    :param thread: 线程对象
    :return:
        True:   关闭成功
        False:  关闭失败
    '''
    try:
        exctype = SystemExit
        tid = ctypes.c_long(thread.ident)
        if not inspect.isclass(exctype):
            exctype = type(exctype)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
        if res == 1:
            return True
        else:
            return False
    except:
        return None


# 线程装饰器
def thread_wrapper(func):
    def wrapper(*args, **kwargs):
        thread_target = Thread(target=func, args=args, kwargs=kwargs)
        thread_target.start()
        return thread_target

    return wrapper



from typing import Literal, Union
from paux import exception
import re


# 计算手续费
def get_commission_data(
        posSide: Literal['long', 'short', 'LONG', 'SHORT'],
        openMoney: Union[int, float],
        openPrice: Union[int, float],
        closePrice: Union[int, float],
        lever: Union[int, float],
        openCommissionRate: Union[int, float],
        closeCommissionRate: Union[int, float]
) -> dict:
    '''
    :param posSide: 持仓方向
        LONG:   多单
        SHORT:  空单
    :param openMoney: 开仓金额
    :param openPrice: 开仓价格
    :param closePrice: 平仓价格
    :param lever: 杠杆倍数
    :param openCommissionRate: 开仓手续费率
    :param closeCommissionRate: 平仓手续费率
    :return
        {
            'closeMoney' : <平仓剩余金额>, # 保留4个小数位
            'commission' : <开仓与平仓手续费之和>, # 保留4个小数位
            'profitRate' : <利润率>, # 保留4个小数位
        }
    '''
    if posSide.upper() == 'LONG':
        buyCommission = openMoney * lever * openCommissionRate
        sellCommission = (closePrice / openPrice) * openMoney * lever * closeCommissionRate
        commission = round(buyCommission + sellCommission, 4)
        closeMoney = round((closePrice - openPrice) / openPrice * lever * openMoney + openMoney - commission, 4)
        profitRate = round((closeMoney - openMoney) / openMoney, 4)
        return dict(
            closeMoney=closeMoney,
            commission=commission,
            profitRate=profitRate,
        )
    elif posSide.upper() == 'SHORT':
        buyCommission = openMoney * lever * openCommissionRate
        sellCommission = (2 * openPrice - closePrice) / openPrice * openMoney * lever * closeCommissionRate
        commission = round(buyCommission + sellCommission, 4)
        closeMoney = (openPrice - closePrice) / openPrice * openMoney * lever + openMoney - commission
        profitRate = round((closeMoney - openMoney) / openMoney, 4)
        return dict(
            closeMoney=closeMoney,
            commission=commission,
            profitRate=profitRate,
        )


# 用于模拟交易中的的价格圆整，保证圆整后的价格与圆整前相差不超过0.001%
def round_simulate(price) -> float:
    '''
    :param price: 价格
    '''
    if isinstance(price, int):
        return price
    if price == 0:
        return price
    price_0_00001 = 0.00001 * price
    if price_0_00001 < 1:
        ndigits = len(str(int(1 / price_0_00001)))
        price_round = round(price, ndigits)
    else:
        price_round = round(price, 1)
    if abs((price_round - price) / price) >= 0.00001:
        msg = f'price_round={price_round},price={price}'
        raise exception.ExecuteException(msg)
    else:
        return price_round


# 圆整购买数量
def round_quantity(
        quantity,
        stepSize: str,
        minQty: str = None,
        maxQty: str = None
):
    '''
    :param quantity: 输入数量
    :param stepSize: 购买数量的最小间隔
    :param minQty: 数量下限, 最小数量   None:无下限
    :param maxQty: 数量上限, 最大数量   None:无上限
    :return
        格式:
            {
                'code' : <int>,
                'data' : <int|float:quantity>,
                'msg' : <str>
            }
        code:
            0     属于 minQty~maxQty之间
            -1    小于minQty
            -2    大于maxQty
    '''

    # 数字对象
    if quantity / float(stepSize) == int(quantity / float(stepSize)):
        quantity = quantity
    else:
        quantity = quantity // float(stepSize) * float(
            stepSize)  # quantity = 52.123 stepSize = 0.01 --> quantity = 52.1
    if '.' in stepSize:
        stepSize = re.sub('0+$', '', stepSize)
        m = len(stepSize.split('.', maxsplit=1)[-1])  # 小数位
        quantity = round(quantity, m)
    else:
        quantity = int(quantity)
    result = {
        'code': 0,
        'data': quantity,
        'msg': ''
    }
    if minQty and quantity < float(minQty):
        result['code'] = -1
        result['msg'] = f'quantity={quantity} minQty={minQty}'
    elif maxQty and quantity > float(maxQty):
        result['code'] = -2
        result['msg'] = f'quantity={quantity} maxQty={maxQty}'
    return result


# 圆整购买价格
def round_price(
        price,
        type: Literal['CEIL', 'FLOOR', 'ceil', 'floor'],
        tickSize,
        minPrice: str = None,
        maxPrice: str = None,

):
    '''
    :param price: 购买价格
    :param type: 圆整方式
        CEIL:   向上圆整
        FLOOR:  向下圆整
    :param tickSize: 价格的最小间隔
    :param minPrice: 最小价格
    :param maxPrice: 最大价格
    :return:
        格式：
            {
                'code' : <int>,
                'data' : <int|float:quantity>,
                'msg' : <str>
            }
        code:
            0     属于 minPrice~maxPrice之间
            -1    小于minPrice
            -2    大于maxPrice
    '''
    # 向下或向上圆整
    type = type.upper()
    if type not in ['CEIL', 'FLOOR']:
        msg = 'type must in ["CEIL","FLOOR"]'
        raise exception.ParamException(msg)
    if price / float(tickSize) == int(price / float(tickSize)):
        price = price
    else:
        if type == 'CEIL':
            price = (price // float(tickSize) + 1) * float(tickSize)  # price = 52.123 tickSize = 0.01 --> price = 52.2
        else:
            price = price // float(tickSize) * float(tickSize)  # price = 52.123 tickSize = 0.01 --> price = 52.1

    if '.' in tickSize:
        tickSize = re.sub('0+$', '', tickSize)
        m = len(tickSize.split('.', maxsplit=1)[-1])  # 小数位
        price = round(price, m)
    else:
        price = int(price)

    result = {
        'code': 0,
        'data': price,
        'msg': ''
    }
    if minPrice and price < float(minPrice):
        result['code'] = -1
        result['msg'] = f'price={price} minPrice={minPrice}'
    elif maxPrice and price > float(maxPrice):
        result['code'] = -2
        result['msg'] = f'price={price} maxPrice={maxPrice}'
    return result


# 根据开仓金额、开仓价格、杠杆获取可以开仓的数量
def get_quantity(
        openPrice: Union[int, float],
        openMoney: Union[int, float],
        stepSize: str,
        lever: int = 1,
        minQty: str = None,
        maxQty: str = None
):
    '''
    :param openPrice: 开仓价格
    :param openMoney: 开仓金额
    :param stepSize: 购买数量的最小间隔
    :param lever: 杠杆倍数
    :param minQty: 数量下限, 最小数量   None:无下限
    :param maxQty: 数量上限, 最大数量   None:无上限
    :return: 结果同round_quantity
    '''
    quantity = openMoney * lever / openPrice
    result = round_quantity(
        quantity=quantity,
        stepSize=stepSize,
        minQty=minQty,
        maxQty=maxQty,
    )
    return result


# 将开仓数量转化为字符串
def quantity_to_f(
        quantity: Union[int, float],
        stepSize: str
):
    '''
    :param quantity: 开仓数量
    :param stepSize: 购买数量的最小间隔
    :return:
        格式：
            {
                'code' : <int>,
                'data' : <str:quantity_f>,
                'msg' : <str>
            }
        code:
            0   正确
            -1  转换后字符串格式quantity与数字格式quantity不相等
    '''
    # m 小数位
    if float(stepSize) >= 1:
        m = 0
    else:
        stepSize = re.sub('0+$', '', stepSize)
        m = len(stepSize.split('.', maxsplit=1)[-1])
    d_format = '%.{m}f'.format(m=m)
    quantity_f = d_format % quantity
    if float(quantity_f) != quantity:
        result = {
            'code': -1,
            'data': quantity_f,
            'msg': f'quantity_f != quantity quantity = {quantity} quantity_f = {quantity_f}',
        }
    else:
        result = {
            'code': 0,
            'data': quantity_f,
            'msg': ''
        }
    return result


# 将开仓价格转化为字符串
def price_to_f(
        price: Union[int, float],
        tickSize: str
):
    '''
    :param quantity: 开仓价格
    :param tickSize: 价格的最小间隔
    :return:
        格式：
            {
                'code' : <int>,
                'data' : <str:price_f>,
                'msg' : <str>
            }
        code:
            0   正确
            -1  转换后字符串格式price与数字格式price不相等
    '''
    # m 小数位
    if float(tickSize) >= 1:
        m = 0
    else:
        tickSize = re.sub('0+$', '', tickSize)
        m = len(tickSize.split('.', maxsplit=1)[-1])
    d_format = '%.{m}f'.format(m=m)
    price_f = d_format % price
    if float(price_f) != price:
        result = {
            'code': -1,
            'data': price_f,
            'msg': f'price_f != price price= {price} price_f= {price_f}',
        }
    else:
        result = {
            'code': 0,
            'data': price_f,
            'msg': ''
        }
    return result



import logging
import os
import time
import traceback
import paux.date as _date


class Log():
    LOG_DIRPATH: str  # './LOG_DATA' 日志文件夹
    FILE_LEVEL: str  # 'INFO' 文件输出级别
    CONSOLE_LEVEL: str  # 'INFO' 控制台打印级别
    SPLIT_LEVEL:bool # 不同级别信息分文件存储
    TIMEZONE:str # 时区


    def __init__(
            self,
            log_dirpath='./LOG_DATA',
            timezone = 'Asia/Shanghai',
            file_level='INFO',
            console_level='DEBUG',
            split_level=True,
    ):
        '''
        :param log_dirpath: 日志文件夹
        :param file_level: 文件输出级别
        :param console_level: 控制台打印级别
        :param split_level: 不同级别信息分文件存储
        '''
        self.LOG_DIRPATH = log_dirpath
        self.FILE_LEVEL = file_level
        self.CONSOLE_LEVEL = console_level
        self.SPLIT_LEVEL = split_level
        self.TIMEZONE = timezone

        if not os.path.isdir(self.LOG_DIRPATH):
            os.makedirs(self.LOG_DIRPATH)

    def get_logger(self, level):
        if not os.path.isdir(self.LOG_DIRPATH):
            os.makedirs(self.LOG_DIRPATH)
        if self.TIMEZONE == 'America/New_York':
            date_str = _date.to_fmt(time.time()*1000,timezone=self.TIMEZONE,fmt='%d-%m-%Y')
        else:
            date_str = _date.to_fmt(time.time()*1000,timezone=self.TIMEZONE,fmt='%Y-%m-%d')
        if self.SPLIT_LEVEL:
            log_filename = '{date}_{level}.log'.format(
                date = date_str,
                level = level.upper(),
            )
        else:
            log_filename = '{date}.log'.format(
                date = date_str,
            )
        log_filepath = os.path.join(self.LOG_DIRPATH, log_filename)
        file_handler = logging.FileHandler(log_filepath)  # 输出到文件
        console_handler = logging.StreamHandler()  # 输出到控制台
        file_handler.setLevel(self.FILE_LEVEL)  # FILE_LEVEL以上才输出到文件
        console_handler.setLevel(self.CONSOLE_LEVEL)  # CONSOLE_LEVEL以上才输出到控制台
        if traceback.format_exc().strip() != 'NoneType: None':
            fmt = '%(asctime)s [%(levelname)s] {message}\ntraceback:{traceback}'.format(filepath=__file__, message='%(message)s', traceback=str(traceback.format_exc()))
        else:
            fmt = '%(asctime)s [%(levelname)s] {message}'.format(filepath=__file__, message='%(message)s')
        formatter = logging.Formatter(fmt)
        file_handler.setFormatter(formatter)  # 设置文件内容格式
        console_handler.setFormatter(formatter)  # 设置控制台内容格式
        logger = logging.getLogger('updateSecurity')
        logger.setLevel(level)
        logger.addHandler(file_handler)
        logger.addHandler(console_handler)
        return logger, [file_handler, console_handler]

    def debug(self, msg):
        LEVEL = 'DEBUG'
        logger, handlers = self.get_logger(level=LEVEL)
        logger.debug(msg)
        for handler in handlers:
            logger.removeHandler(handler)

    def info(self, msg):
        LEVEL = 'INFO'
        logger, handlers = self.get_logger(level=LEVEL)
        logger.info(msg)
        for handler in handlers:
            logger.removeHandler(handler)

    def error(self, msg):
        LEVEL = 'ERROR'
        logger, handlers = self.get_logger(level=LEVEL)
        logger.error(msg)
        for handler in handlers:
            logger.removeHandler(handler)

    def warn(self, msg):
        LEVEL = 'WARN'
        logger, handlers = self.get_logger(level=LEVEL)
        logger.warning(msg)
        for handler in handlers:
            logger.removeHandler(handler)

    def critical(self, msg):
        LEVEL = 'CRITICAL'
        logger, handlers = self.get_logger(level=LEVEL)
        logger.critical(msg)
        for handler in handlers:
            logger.removeHandler(handler)


from multiprocessing import Process, Manager
import traceback
import time
from paux import exception


# 进程工作者
def _pool_worker(q_param, q_result):
    '''
    :param q_param: 参数队列
    :param q_result: 结果队列
    '''
    while True:
        try:
            data = q_param.get(block=False, timeout=0)
            index = data['index']  # 索引
            param = data['param']  # 函数的执行参数
            skip_exception = data['skip_exception']  # 遇到异常不会终止工作器
            func = param['func']  # 函数
            del param['func']
        except:
            break
        if skip_exception:
            try:
                ret = func(**param)
                q_result.put(
                    {
                        'index': index,
                        'data': ret
                    }
                )
            except:
                print(traceback.format_exc())
        else:
            ret = func(**param)
            q_result.put(
                {
                    'index': index,
                    'data': ret
                }
            )


# 进程装饰器
def process_wrapper(func):
    def wrapper(*args, **kwargs):
        process_target = Process(target=func, args=args, kwargs=kwargs)
        process_target.run()
        return process_target

    return wrapper


def pool_worker(
        params,
        p_num=4,
        func=None,
        skip_exception=False
):
    '''
    :param params: 参数序列 [dict,dict....]
    :param p_num: 进程数
    :param func: 执行函数
        函数地址如果在params中，func可以为None
        优先级: params[<index>]['func'] >> func
    :param skip_exception: 子进程中出现异常是否终止
        True:   出现异常，报告错误信息，不会终止子进程
        False:  出现异常，终止子进程
    :return
        [result、result... ...]
    '''
    # 结果按照params的索引对应
    results = [None] * len(params)
    # 标准化param，让param中存在要执行的函数func
    q_param = Manager().Queue()  # 参数队列
    q_result = Manager().Queue()  # 结果队列
    for index, param in enumerate(params):
        # param['func']优先级最高
        # 存在    param['func']
        if 'func' in param.keys() and param['func'] != None:
            pass
        # 不存在   param['func']
        else:
            # 有func
            if func != None:
                param['func'] = func
            # 无func
            else:
                msg = 'No func to execute'
                raise exception.ParamException(msg)
        q_param.put(
            {
                'index': index,
                'param': param,
                'skip_exception': skip_exception
            }
        )
    # 单进程运行
    if p_num <= 1:
        _pool_worker(q_param, q_result)
    # 多进程运行
    else:
        processes = []
        for i in range(p_num):
            p = Process(target=_pool_worker, kwargs={'q_param': q_param, 'q_result': q_result})
            processes.append(p)
            p.start()

        def wait_processes(processes):
            for p in processes:
                if p.is_alive():
                    return False
            return True

        # 等待进程均运行完成
        while True:
            if wait_processes(processes):
                break
            else:
                time.sleep(0.15)
    # 整理结果
    for i in range(q_result.qsize()):
        result = q_result.get(block=False, timeout=0)
        index = result['index']
        data = result['data']
        # 按照索引赋值，如果某个参数执行异常并且skip_exception，这个结果为None
        results[index] = data
    return results



from typing import Union
import time


class Filter():
    def __init__(self):
        # 过滤内容，格式：
        # {
        #     '<name>':{
        #         'ts':<int>,
        #         'expire':<int>,
        #     },
        # }
        # 时间戳以秒为单位
        self.filter_map = {}

    # 设置过滤对象
    def set(
            self,
            name,
            filter_minute: Union[int, float] = 5
    ):
        '''
        :param name: 对象名称
        :param filter_minute: 缓存的分钟数
        '''
        ts = int(time.time())
        self.filter_map[name] = {
            'ts': ts,
            'expire': ts + filter_minute * 60
        }

    # 检查是否被过滤
    def check(self, name):
        '''
        :param name: 对象名称
        :return:
            True 不过滤
            False 过滤
        '''

        # 清除超过缓存的数据
        self.clear()
        # filter_map中仅保留未过期的缓存对象
        if not name in self.filter_map.keys():
            return True  # 不过滤
        else:
            return False  # 过滤

    # 清除过期的过滤对象
    def clear(self):
        this_time = time.time()
        remove_keys = []
        for key, data in self.filter_map.items():
            if this_time >= data['expire']:
                remove_keys.append(key)
        for key in remove_keys:
            del self.filter_map[key]
